from ctypes import c_char_p, get_errno
from errno import errorcode
from os import strerror

from typing import Tuple


def str_to_char_p(value: str) -> bytes:
    """Creates a char * from Python string.

    :param value: Python string

    """
    return c_char_p(value.encode('utf-8'))


def str_from_char_p(value: bytes) -> str:
    """Creates Python string from char *.

    :param value: char * value

    """
    return c_char_p(value).value.decode('utf-8')


def get_last_error() -> Tuple[int, str, str]:
    """Returns last error (``errno``) information tuple:

        (err_no, err_code, err_message)

    """
    num = get_errno()
    code = errorcode[num]
    msg = strerror(num)

    return num, code, msg



import ctypes
import inspect
import logging
import os
from collections import namedtuple, OrderedDict
from contextlib import contextmanager
from ctypes.util import find_library
from functools import partial, partialmethod
from typing import Any, Optional, Callable

from .exceptions import FunctionRedeclared, UnsupportedTypeError, TypehintError
from .types import CStr
from .utils import str_from_char_p

LOGGER = logging.getLogger(__name__)


FuncInfo = namedtuple('FuncInfo', ['name_py', 'name_c', 'annotations'])


_MISSING = namedtuple('MissingType', [])


class Library:
    """Main entry point to describe C library interface.

    Basic usage:

    .. code-block:: python

        lib = Library('mylib')

        with lib.functions_prefix('mylib_'):

            @lib.function()
            def my_func():
                ...

        lib.bind_types()

    """

    def __init__(self, name: str, *, autoload: bool=True):
        """

        :param name: Shared library name or filepath.
        :param autoload: Load library just on Library object initialization.
        """
        self.name = name
        self.lib = None
        self.funcs = {}
        self._prefixes = []

        autoload and self.load()

    @contextmanager
    def functions_prefix(self, prefix: str):
        """Nestable context manager. Used for namespacing.

        Useful when C functions have common prefixes.

        .. code-block:: python

            with lib.functions_prefix('MyLib_'):

                @lib.f('my_func')  # Will be expanded into `MyLib_my_func`
                def some_func():
                    ...

        :param prefix: Function name prefix to apply to

        """
        prefixes = self._prefixes

        prefixes.append(prefix)

        yield

        prefixes.pop()

    def load(self):
        """Loads shared library."""
        name = self.name

        if not os.path.exists(name):
            name = find_library(name)

        lib = ctypes.CDLL(name, use_errno=True)

        if lib._name is None:
            lib = None

        self.lib = lib

    def function(self, name_c: Optional[str]=None, *, wrap: bool=False):
        """Decorator to mark functions which exported from the library.

        :param name_c: C function name with or without prefix (see ``.functions_prefix()``).
            If not set, Python function name is used.

        :param wrap: Do not replace decorated function with ctypes function,
            but with wrapper, allowing pre- or post-process ctypes function call.

            Useful to organize functions to classes (to automatically pass ``self``)
            to ctypes function to C function.

            .. code-block:: python

                class Thing(CObject):

                    @lib.function(wrap=True)
                    def one(self, some: int) -> int:
                        # Implicitly pass Thing instance alongside
                        # with explicitly passed `some` arg.
                        ...

                    @lib.function(wrap=True)
                    def two(self, some:int, cfunc: Callable) -> int:
                        # `cfunc` is a wrapper, calling an actual ctypes function.
                        # If no arguments provided the wrapper will try detect them
                        # automatically.
                        result = cfunc()
                        return result + 1


        """
        def cfunc_wrapped(*args, f: Callable, **kwargs):

            if not args:
                argvals = inspect.getargvalues(inspect.currentframe().f_back)
                loc = argvals.locals
                args = [loc[argname] for argname in argvals.args if argname != 'cfunc']

            return f(*args)

        def cfunc_direct(*args, f: Callable, **kwargs):
            return f(*args)

        def function_(func_py: Callable):

            info = self._extract_func_info(func_py, name_c=name_c)
            name = info.name_c

            func_c = getattr(self.lib, name)

            # Prepare for late binding in .bind_types().
            func_c.ctyped = info

            if wrap:
                func_args = inspect.getfullargspec(func_py).args

                if 'cfunc' in func_args:
                    # Use existing function, pass `cfunc`.

                    LOGGER.debug(
                        'Function %s (%s) is wrapped and implements manual `cfunc` handling.', info.name_py, name)

                    func_swapped = partialmethod(func_py, cfunc=partial(cfunc_wrapped, f=func_c))

                else:
                    # Use automatic function.

                    LOGGER.debug(
                        'Function %s (%s) is wrapped and implements automatic proxied ctypes call.', info.name_py, name)

                    func_swapped = partialmethod(cfunc_direct, f=func_c)

                func_swapped.cfunc = func_c
                func_out = func_swapped

            else:

                LOGGER.debug('Function %s is swapped with direct ctypes %s call.' % (info.name_py, name))

                func_out = func_c

            self.funcs[name] = func_out

            return func_out

        return function_

    def method(self, name_c: Optional[str]=None):
        """Decorator. The same as ``.function()`` with ``wrap=True``."""
        return self.function(name_c=name_c, wrap=True)

    def bind_types(self):
        """Deduces ctypes argument and result types from Python typehints,
        binding those types to ctypes functions.


        """
        LOGGER.debug('Binding signature types to ctypes functions ...')

        def thint_str_to_obj(thint: str):
            fback = inspect.currentframe().f_back

            while fback:
                target = fback.f_globals.get(thint)

                if target:
                    return target

                fback = fback.f_back

        def cast_type(name: str, thint: Any):

            if thint is None:
                return None

            if isinstance(thint, str):
                thint_orig = thint
                thint = thint_str_to_obj(thint)

                if thint is None:
                    raise TypehintError(
                        'Unable to resolve typehint. Function: %s. Arg: %s. Type: %s. ' %
                        (name_py, name, thint_orig))

            if name == 'return' and thint is str:
                # Cast result to Python string.
                return str_from_char_p

            if thint is str:
                thint = CStr

            return thint

        for name_c, func_out in self.funcs.items():

            func_c = getattr(func_out, 'cfunc', func_out)
            func_info: FuncInfo = func_c.ctyped

            name_py = func_info.name_py
            annotations = func_info.annotations

            try:
                restype = cast_type('return', annotations.pop('return', None))
                argtypes = [cast_type(argname, argtype) for argname, argtype in annotations.items()]

            except TypehintError:
                # Reset annotations to allow subsequent .bind_types() calls w/o exceptions.
                func_info.annotations.clear()
                raise

            try:
                if argtypes:
                    func_c.argtypes = argtypes

                if restype:
                    func_c.restype = restype

            except TypeError as e:

                raise UnsupportedTypeError(
                    'Unsupported types declared for %s (%s). Args: %s. Result: %s.' %
                    (name_py, name_c, argtypes, restype)) from e

    #####################################################################################
    # Shortcuts

    f = function
    m = method

    #####################################################################################
    # Private

    def _extract_func_info(self, func: Callable, *, name_c: Optional[str]=None) -> FuncInfo:

        name_py = func.__name__
        name = name_c or name_py
        name = ''.join(self._prefixes) + name

        if name in self.funcs:
            raise FunctionRedeclared('Unable to redeclare: %s (%s)' % (name, name_py))

        annotated_args = OrderedDict()
        annotations = func.__annotations__

        # Gather all args and annotations for them.
        for argname in inspect.getfullargspec(func).args:

            if argname == 'cfunc':
                continue

            annotation = annotations.get(argname, _MISSING)

            if argname == 'self' and annotation is _MISSING:
                # Pick class name from qualname.
                annotation = func.__qualname__.split('.')[-2]

            annotated_args[argname] = annotation

        annotated_args['return'] = annotations.get('return')

        return FuncInfo(name_py=name_py, name_c=name, annotations=annotated_args)





class CtypedException(Exception):
    """"""


class FunctionRedeclared(CtypedException):
    """"""


class UnsupportedTypeError(CtypedException):
    """"""


class TypehintError(CtypedException):
    """"""



import os

import pytest

from ctyped.exceptions import FunctionRedeclared, TypehintError, UnsupportedTypeError
from ctyped.toolbox import Library, get_last_error
from ctyped.types import CInt

############################################################
# Library interface

MYLIB_PATH = os.path.join(os.path.dirname(__file__), 'mylib', 'mylib.so')

mylib = Library(MYLIB_PATH)


@mylib.f()
def f_noprefix_1() -> int:
    ...


@mylib.f()
def with_errno() -> int:
    ...


with mylib.functions_prefix('f_prefix_one_'):

    @mylib.function('func_1')
    def function_one() -> int:
        ...

    @mylib.function()
    def func_2() -> int:
        ...

    @mylib.f('char_p')
    def func_str(some: str) -> str:
        ...

    with mylib.functions_prefix('prefix_two_'):

        @mylib.f('func_3')
        def func_prefix_two_3() -> int:
            ...

    class Prober(CInt):

        @mylib.m()
        def probe_add_one(self) -> int:
            ...

        @mylib.m('probe_add_two')
        def probe_add_three(self, cfunc) -> int:
            result = cfunc()
            return result + 1

    @mylib.f()
    def get_prober() -> Prober:
        ...


mylib.bind_types()

############################################################
# Tests


def test_basic():
    assert f_noprefix_1() == -10
    assert function_one() == 1
    assert func_2() == 2
    assert func_prefix_two_3() == 3

    prober = get_prober()
    assert isinstance(prober, Prober)

    prober_val = prober.value

    assert prober.probe_add_one() == prober_val + 1
    assert prober.probe_add_three() == prober_val + 3


def test_with_errno():
    assert with_errno() == 333
    err_num, err_code, err_msg = get_last_error()
    assert err_num == 2
    assert err_code == 'ENOENT'
    assert 'such file' in err_msg


def test_strings():

    assert func_str('mind') == 'hereyouare: mind'


def test_no_redeclare():

    with pytest.raises(FunctionRedeclared):

        @mylib.f()
        def f_noprefix_1() -> int:
            ...


def test_unresolved_typehint():

    @mylib.f()
    def buggy1(one: 'SomeDummyType') -> int:
        ...

    with pytest.raises(TypehintError) as e:
        mylib.bind_types()

    assert 'SomeDummyType' in str(e.value)


def test_unsupported_type():

    class SomeType: pass

    @mylib.f()
    def buggy2(one: SomeType) -> int:
        ...

    with pytest.raises(UnsupportedTypeError) as e:
        mylib.bind_types()

    assert 'buggy2 (buggy2)' in str(e.value)



import ctypes

from .utils import str_to_char_p


class CObject(ctypes.c_void_p):
    """"""


# getattr to cheat type hints
CShort: int = getattr(ctypes, 'c_short')
CShortU: int = getattr(ctypes, 'c_ushort')
CLong: int = getattr(ctypes, 'c_long')
CLongU: int = getattr(ctypes, 'c_ulong')
CLongLong: int = getattr(ctypes, 'c_longlong')
CLongLongU: int = getattr(ctypes, 'c_ulonglong')

CInt: int = getattr(ctypes, 'c_int')
CIntU: int = getattr(ctypes, 'c_uint')
CInt8: int = getattr(ctypes, 'c_int8')
CInt8U: int = getattr(ctypes, 'c_uint8')
CInt16: int = getattr(ctypes, 'c_int16')
CInt16U: int = getattr(ctypes, 'c_uint16')
CInt32: int = getattr(ctypes, 'c_int32')
CInt32U: int = getattr(ctypes, 'c_uint32')
CInt64: int = getattr(ctypes, 'c_int64')
CInt64U: int = getattr(ctypes, 'c_uint64')

CPointer: CObject = getattr(ctypes, 'c_void_p')


class CStr:
    """Represents Python string as a C chars pointer."""

    @classmethod
    def from_param(cls, val):
        return str_to_char_p(val)
